#pragma once
#include <SFML/Graphics.hpp>
#include <unordered_map>
#include <string>
#include <fstream>
namespace zle
{
	static const std::string ZLE_VERSION = "#1.0";
	static const int ZLE_blendCNT = 4;
	static const sf::BlendMode blendModes[ZLE_blendCNT] = { sf::BlendMode(sf::BlendMode::One, sf::BlendMode::Zero),
	sf::BlendMode(sf::BlendMode::SrcAlpha, sf::BlendMode::One, sf::BlendMode::Add, sf::BlendMode::One, sf::BlendMode::One, sf::BlendMode::Add),
	sf::BlendMode(sf::BlendMode::SrcAlpha, sf::BlendMode::OneMinusSrcAlpha, sf::BlendMode::Add, sf::BlendMode::One, sf::BlendMode::OneMinusSrcAlpha, sf::BlendMode::Add),
	sf::BlendMode(sf::BlendMode::DstColor, sf::BlendMode::Zero) };
	class TileMap : public sf::Transformable, public sf::Drawable
	{
		sf::VertexArray arr;
		sf::VertexBuffer buff;
		sf::Vector2u mapSize;
		sf::Vector2u tileSize;
		sf::Vector2u textureTileSize;
		sf::Color overAllColor;
		sf::Texture* texture;
		unsigned int textureGap;
		std::vector<std::vector<unsigned short>> mapData;
		void InternalUpdate()
		{
			//resize buffer
			mapData.resize(mapSize.x);
			for (int i = 0; i < mapSize.x; i++)
				mapData[i].resize(mapSize.y);
			arr.setPrimitiveType(sf::Triangles);
			arr.resize(6 * mapSize.x * mapSize.y);
			buff.setPrimitiveType(sf::Triangles);
			buff.setUsage(sf::VertexBuffer::Dynamic);

			//setup data
			for (int i = 0; i < mapSize.x; i++)
				for (int j = 0; j < mapSize.y; j++)
				{
					sf::Vector2f position = sf::Vector2f(tileSize.x * i, tileSize.y * j);
					unsigned int v = (j + mapSize.y * i) * 6;

					arr[v].position = position;
					arr[v].color = overAllColor;
					arr[v].texCoords = sf::Vector2f(mapData[i][j] * textureTileSize.x + textureGap * mapData[i][j], 0);

					arr[v + 1].position = position + sf::Vector2f(tileSize.x, 0);
					arr[v + 1].color = overAllColor;
					arr[v + 1].texCoords = sf::Vector2f(mapData[i][j] * textureTileSize.x + textureTileSize.x + textureGap * mapData[i][j], 0);

					arr[v + 2].position = position + (sf::Vector2f)tileSize;
					arr[v + 2].color = overAllColor;
					arr[v + 2].texCoords = sf::Vector2f(mapData[i][j] * textureTileSize.x + textureTileSize.x + textureGap * mapData[i][j], textureTileSize.y);

					arr[v + 3].position = position + sf::Vector2f(0, tileSize.y);
					arr[v + 3].color = overAllColor;
					arr[v + 3].texCoords = sf::Vector2f(mapData[i][j] * textureTileSize.x + textureGap * mapData[i][j], textureTileSize.y);

					arr[v + 4].position = position;
					arr[v + 4].color = overAllColor;
					arr[v + 4].texCoords = sf::Vector2f(mapData[i][j] * textureTileSize.x + textureGap * mapData[i][j], 0);

					arr[v + 5].position = position + (sf::Vector2f)tileSize;
					arr[v + 5].color = overAllColor;
					arr[v + 5].texCoords = sf::Vector2f(mapData[i][j] * textureTileSize.x + textureTileSize.x + textureGap * mapData[i][j], textureTileSize.y);
				}

			//update
			buff.create(arr.getVertexCount());
			buff.update(&arr[0]);
		}
	public:
		bool drawWithoutTransform = 0;
		TileMap()
		{
			overAllColor = sf::Color::White;
			mapSize = sf::Vector2u(1, 1);
			tileSize = sf::Vector2u(1, 1);
		}
		~TileMap() {}
		void setTileCount(sf::Vector2u mapSize)
		{
			this->mapSize = mapSize;
			InternalUpdate();
		}
		sf::Vector2u getTileCount() const
		{
			return mapSize;
		}
		void setTexture(sf::Texture* texture)
		{
			this->texture = texture;
		}
		const sf::Texture* getTexture() const
		{
			return texture;
		}
		void setTileSize(sf::Vector2u tileSize)
		{
			this->tileSize = tileSize;
			InternalUpdate();
		}
		sf::Vector2u getTileSize() const
		{
			return tileSize;
		}
		void setTextureTileSize(sf::Vector2u textureTileSize)
		{
			this->textureTileSize = textureTileSize;
			InternalUpdate();
		}
		void setTextureGap(unsigned int gap)
		{
			textureGap = gap;
			InternalUpdate();
		}
		unsigned int getTextureGap() const
		{
			return textureGap;
		}
		sf::Vector2u getTextureTileSize() const
		{
			return textureTileSize;
		}
		void setColor(sf::Color color)
		{
			overAllColor = color;
			for (int i = 0; i < mapSize.x; i++)
				for (int j = 0; j < mapSize.y; j++)
				{
					unsigned int v = (j + mapSize.y * i) * 6;

					arr[v].color = overAllColor;
					arr[v + 1].color = overAllColor;
					arr[v + 2].color = overAllColor;
					arr[v + 3].color = overAllColor;
					arr[v + 4].color = overAllColor;
					arr[v + 5].color = overAllColor;
				}
			buff.update(&arr[0]);
		}
		sf::Color getColor() const
		{
			return overAllColor;
		}
		unsigned short getTileID(sf::Vector2u tile) const
		{
			return mapData[tile.x][tile.y];
		}
		void setTileID(sf::Vector2u tile, unsigned short ID)
		{
			mapData[tile.x][tile.y] = ID;

			unsigned int v = (tile.y + mapSize.y * tile.x) * 6;
			arr[v].texCoords = sf::Vector2f(ID * textureTileSize.x + textureGap * ID, 0);
			arr[v + 1].texCoords = sf::Vector2f(ID * textureTileSize.x + textureGap * ID + textureTileSize.x, 0);
			arr[v + 2].texCoords = sf::Vector2f(ID * textureTileSize.x + textureGap * ID + textureTileSize.x, textureTileSize.y);
			arr[v + 3].texCoords = sf::Vector2f(ID * textureTileSize.x + textureGap * ID, textureTileSize.y);
			arr[v + 4].texCoords = sf::Vector2f(ID * textureTileSize.x + textureGap * ID, 0);
			arr[v + 5].texCoords = sf::Vector2f(ID * textureTileSize.x + textureGap * ID + textureTileSize.x, textureTileSize.y);
			buff.update(&arr[0]);
		}
	private:
		virtual void draw(sf::RenderTarget& target, sf::RenderStates states) const
		{
			if (!drawWithoutTransform)
				states.transform *= getTransform();
			states.texture = texture;
			if (sf::VertexBuffer::isAvailable())
				target.draw(buff, states);
			else
				target.draw(arr, states);
		}
	};
	class Level : public sf::Drawable, public sf::Transformable
	{
		enum class ObjectType
		{
			None,
			Tilemap,
			Sprite,
			Texture,

		};
		struct Object : public sf::Drawable
		{
			sf::Sprite* sprite;
			TileMap* tilemap;
			sf::Texture* usingTexture;
			ObjectType type;
			sf::Int8 drawingZOrder;
			bool selected = 0;
			bool visible = 1;
			sf::BlendMode blendMode;
			sf::Color color;
			Object()
				: drawingZOrder(0)/*, type(ObjectType::None)*/
			{
				sprite = nullptr;
				tilemap = nullptr;
				blendMode = sf::BlendAlpha;
			}
			~Object()
			{
				if (sprite)
					delete sprite;
				if (tilemap)
					delete tilemap;
			}
			bool UseSprite()
			{
				if (tilemap || sprite)
					return 0;
				sprite = new sf::Sprite;
				return 1;
			}
			bool UseTileMap()
			{
				if (tilemap || sprite)
					return 0;
				tilemap = new TileMap;
				return 1;
			}
			void* getPointer() const
			{
				if (tilemap)
					return tilemap;
				if (sprite)
					return sprite;
			}
			sf::Transformable& getTransformable() const
			{
				if (tilemap)
					return (*tilemap);
				if (sprite)
					return (*sprite);
				return (*sprite);
			}
		private:
			virtual void draw(sf::RenderTarget& target, sf::RenderStates states) const
			{
				states.blendMode = blendMode;
				if (sprite)
					target.draw(*sprite, states);
				if (tilemap)
					target.draw(*tilemap, states);
			}
		};
		struct Tobject
		{
			sf::Texture texture;
			std::string textureLocation;
			bool selected;
			Tobject()
			{

			}
			~Tobject()
			{

			}
		};
		std::unordered_map <std::string, Object> objects;
		std::unordered_map <std::string, Tobject> texture_objects;
	public:
		Level()
		{

		}
		~Level()
		{

		}
		bool loadFromFile(const std::string& fileName, std::string resourceFolder)
		{
			std::ifstream load;
			load.open(fileName);
			if (load.is_open())
			{

				std::string line;
				load >> line;
				if (line != ZLE_VERSION)
					return 0;
				load >> line;
				int counter;
				int temp;
				float tempF;
				load >> counter;
				std::vector<sf::Texture*> textureP;
				textureP.resize(counter);
				for (int i = 0; i < counter; i++)
				{
					load >> line;
					std::string line2 = line;
					int count = 0;
					while (1)
					{
						if (line2[0] == '\\' || line2[0] == '/')
						{
							line2.erase(0, 1);
							line.erase(0, 1);
							count++;
							if (count == 2)
							{
								if (resourceFolder[resourceFolder.size() - 1] != '\\' && resourceFolder[resourceFolder.size() - 1] != '/')
									resourceFolder.push_back('/');
								line.insert(0, resourceFolder);
								break;
							}
						}
						else
						{
							line2.erase(0, 1);
							line.erase(0, 1);
						}

					}
					texture_objects[line2].textureLocation = line;
					texture_objects[line2].texture.loadFromFile(line);
					load >> temp;
					texture_objects[line2].texture.setSmooth(temp);
					load >> temp;
					texture_objects[line2].texture.setRepeated(temp);
					texture_objects[line2].selected = 0;
					textureP[i] = &texture_objects[line2].texture;
				}
				load >> line;
				load >> counter;
				for (int i = 0; i < counter; i++)
				{
					load >> line;
					load >> temp;
					Object& o = objects[line];
					o.color.r = temp;
					load >> temp;
					o.color.g = temp;
					load >> temp;
					o.color.b = temp;
					load >> temp;
					o.color.a = temp;
					load >> temp;
					o.type = (ObjectType)temp;
					switch (o.type)
					{
					case ObjectType::Sprite:
						o.UseSprite();
						break;
					case ObjectType::Tilemap:
						o.UseTileMap();
						break;
					}
					load >> temp;
					o.usingTexture = textureP[temp];
					load >> temp;
					o.visible = temp;
					load >> temp;
					o.drawingZOrder = temp;
					load >> temp;
					o.blendMode = blendModes[temp];

					sf::Vector2f tempV;
					sf::Transformable& t = o.getTransformable();
					load >> tempV.x;
					load >> tempV.y;
					t.setPosition(tempV);

					load >> tempF;
					t.setRotation(tempF);

					load >> tempV.x;
					load >> tempV.y;
					t.setScale(tempV);

					load >> tempV.x;
					load >> tempV.y;
					t.setOrigin(tempV);

					switch (o.type)
					{
					case ObjectType::Sprite:
					{
						sf::Sprite* s = o.sprite;
						s->setColor(o.color);

						sf::IntRect tempI;
						load >> tempI.left;
						load >> tempI.top;
						load >> tempI.width;
						load >> tempI.height;
						s->setTextureRect(tempI);

						s->setTexture(*o.usingTexture);
					}
					break;
					case ObjectType::Tilemap:
					{
						TileMap* s = o.tilemap;


						load >> temp;
						s->setTextureGap(temp);

						load >> tempV.x;
						load >> tempV.y;
						s->setTextureTileSize((sf::Vector2u)tempV);
						load >> tempV.x;
						load >> tempV.y;
						s->setTileCount((sf::Vector2u)tempV);
						load >> tempV.x;
						load >> tempV.y;
						s->setTileSize((sf::Vector2u)tempV);

						s->setColor(o.color);

						for (int i = 0; i < s->getTileCount().x; i++)
							for (int j = 0; j < s->getTileCount().y; j++)
							{
								load >> temp;
								s->setTileID(sf::Vector2u(i, j), temp);
							}
						s->setTexture(o.usingTexture);
					}
					break;
					}

				}
				load.close();
				return 1;
			}
			return 0;
		}


	private:
		virtual void draw(sf::RenderTarget& target, sf::RenderStates states) const
		{
			states.transform *= getTransform();
			for (int i = INT8_MIN; i < INT8_MAX; i++)
			{
				for (auto& n : objects)
				{
					if (!n.second.visible)
						continue;
					if (n.second.drawingZOrder == i)
					{
						target.draw(n.second, states);
					}
				}
			}
		}
	};
}